# file: rarest.tcl

set ns [new Simulator]

# parameter values
set opt(M)    100
set opt(U)    4
set opt(D)    16
set opt(INTVAL) 01.0
set opt(SDRES) 20.0
set opt(STSDCNT) 1
set opt(DLTIME) 5.0
set opt(TRINT) 1.0

#artificial: estimates artificial seeds and download time
set artseedcnt [expr $opt(SDRES)/$opt(INTVAL)]
set opt(artseedcnt) [expr int([expr 2.0 * $artseedcnt])]
set opt(artdltime)  [expr [ expr [expr 2.0 * $opt(DLTIME)] * $opt(M)]/$opt(U)]
set opt(artseedcnt) 0
puts "artificial seedcnt $opt(artseedcnt) artificial dltime $opt(artdltime)"


# statistical counters
set pst(peerno) 0
set pst(free) 0
set pst(demand) 0
set pst(dlcnt) 0
set pst(sdcnt) 0
set pst(activeu) 0

source related.tcl

Class PDATA;

PDATA instproc init {pno} {
   global ns opt

   $self instvar peerno
   set peerno $pno
    
   $self instvar activeu actived downloaded uploaded reqtime progress ptype ualloc dalloc seqdownloaded arrivaltime
   set activeu 0
   set actived 0
   set ualloc 0
   set dalloc 0
   set downloaded 0
   set seqdownloaded 0	
   set uploaded 0
   set progress 0
   set reqtime [$ns now]
   set ptype "D"

   #matrix: keeps the matrix of downloaded piece
   $self instvar dpiece upiece
   for { set i 0 } { $i <= $opt(M) } { incr i } {
	lappend dpiece 0
        lappend upiece 0
   }  

   vputs "peer $peerno : data created"
}



PDATA instproc print {} {
   global ns

   $self instvar downloaded uploaded activeu actived peerno progress

   puts "Peer $peerno downloaded: $downloaded uploaded: $uploaded progress: $progress active: $activeu u  $actived d"
      
}


PDATA instproc sprint {} {
   global ns

   $self instvar downloaded uploaded activeu actived peerno progress

   puts "Peer $peerno data"
   puts "downloaded: $downloaded "
   puts "uploaded: $uploaded "
   puts "progress: $progress"
   puts "active: $activeu u  $actived d"
}



proc theputs {msg} {
    global ns

    return 0
    set output [format "%9.4f  $msg" [$ns now]]
    puts $output

}


proc vputs {msg} {
    global ns

    return 0
    set output [format "%9.4f  $msg" [$ns now]]
    puts $output

}


proc needputs {msg} {
    global ns

    set output [format "%9.4f  $msg" [$ns now]]
    puts $output
    return 0
}


proc seeddepart {uloader} {
   global pst ns opt


   vputs "seed  [$uloader set peerno] depart invoked ( activeu [$uloader set activeu] )"   
   $uloader set ptype    "SD"

   set activeu [$uloader set activeu]
   if { $activeu > 0 } {
	$ns at [expr [$ns now] + 1.0 ] "seeddepart $uloader"
        return 
   } 

   incr pst(sdcnt) -1
   incr pst(free) -$opt(U)
   vputs "seed [$uloader set peerno] departing "

   # deleting the peer
   set dno [$uloader set peerno]
   set count [llength $pst(dlist)]
   set index -1 
   for {set i 0} {$i < $count} { incr i} {
       set tpeer [lindex $pst(dlist) $i]
       if {[$tpeer set peerno]==$dno} {
            set index $i
            break
       }
   }     
   if {$index < 0} {puts "Peerdepart: invalid delete request"; exit 1}
   set pst(dlist) [edelete $pst(dlist) $index]       

}


proc dcomplete {dloader uloader tpiece} {
   global opt ns pst
  
   set activeu [$uloader set activeu]
   incr activeu -1
   $uloader set activeu $activeu
   incr pst(activeu) -1

   set actived [$dloader set actived]
   incr actived -1
   $dloader set actived $actived

   set ualloc [$uloader set ualloc]
   incr ualloc -1
   $uloader set ualloc $ualloc

   set dalloc [$dloader set dalloc]
   incr dalloc -1
   $dloader set dalloc $dalloc

   set downloaded [$dloader set downloaded]
   incr downloaded 1
   $dloader set downloaded $downloaded

   # no download if downloaded >= M
   if { $downloaded == 1 } {
	incr pst(free) $opt(U)
   }

   incr pst(free) 1
   incr pst(demand) 1
   
   if { $downloaded == $opt(M) && $actived == 0 } {
	incr pst(demand) -$opt(D)
        incr pst(dlcnt) -1
        incr pst(sdcnt)  1
        $dloader set ptype "S"
        needputs "peer [$dloader set peerno] becoming seed downloadtime [expr [$ns now] - [$dloader set arrivaltime]]"
        
        set rcomp [expr rand() * $opt(SDRES) ]
        set rcomp [expr $rcomp * 0.25]
        set rcomp [expr $rcomp + $opt(SDRES) ]
#        set rcom $opt(SDRES)
        $ns at [expr [$ns now] + $rcomp ] "seeddepart $dloader"
   }
   

    # increase the trace counter of the piece
    set tpcount [lindex $opt(piececount) $tpiece]
    incr tpcount 1
    set opt(piececount) [lreplace $opt(piececount) $tpiece $tpiece $tpcount]


   # matrix: upload/download matrix 
   set dpiece [$dloader set dpiece]
   set dpiece [lreplace $dpiece $tpiece $tpiece c]
   $dloader set dpiece $dpiece

   set upiece [$uloader set upiece]
   set cnt [lindex $upiece $tpiece]
   incr cnt -1
   set upiece [lreplace $upiece $tpiece $tpiece $cnt]
   $uloader set upiece $upiece
   
   # update seqdownloaded
   set seqdownloaded [$dloader set seqdownloaded]
   set i 1
   if { $seqdownloaded > 0 } { set i $seqdownloaded}	
   for { set j 1 } { $i <= $opt(M) } { incr i} {
	set ds [lindex $dpiece $i]
	if { $ds == "c" } { set seqdownloaded $i }
	if { $ds == "0" } { break }
	if { $ds == "1" } { break }
	if { $ds == "i" } { break }
   }
   $dloader set seqdownloaded $seqdownloaded


   needputs " download complete  d [$dloader set peerno] u [$uloader set peerno] p $tpiece downloaded $downloaded seqdownloaded $seqdownloaded elapsed [expr [$ns now] - [$dloader set arrivaltime]] "

   if { $seqdownloaded > $downloaded } {
	puts "high sequence download"
	puts $dpiece
	exit 655
   }


   if { $pst(free) > 0 && $pst(demand) > 0 } {
        theputs "tracker called by dcomplete"
	tracker	
   }

   if { $downloaded == $opt(M) } { vputs " all download complete $dpiece"}
   if { $downloaded > $opt(M) } { puts "high download"; exit 1}
}


proc startdownload {dloader uloader tpiece} {
   global opt ns pst

   # no upload if activeu > U
   set activeu [$uloader set activeu]
   incr activeu 1
   $uloader set activeu $activeu
   incr pst(activeu) 1
   
   # no download if actived > D
   set actived [$dloader set actived]
   incr actived 1
   $dloader set actived $actived

   # scheduling dcomplete
   set ctime [$ns now]
   $dloader set reqtime $ctime
   set rcomp [expr rand() * $opt(DLTIME) ]
   set rcomp [expr $rcomp * 0.25]
   set rcomp [expr $rcomp + $opt(DLTIME) ]
#   set rcomp $opt(DLTIME)
   $ns at [expr $ctime + $rcomp] "dcomplete $dloader $uloader $tpiece"

   vputs "  download started  d [$dloader set peerno] u [$uloader set peerno] p $tpiece"

   # matrix: download/upload matrix 
   set dpiece [$dloader set dpiece]
   set dpiece [lreplace $dpiece $tpiece $tpiece 1]
   $dloader set dpiece $dpiece

   set upiece [$uloader set upiece]
   set cnt [lindex $upiece $tpiece]
   incr cnt 1 
   set upiece [lreplace $upiece $tpiece $tpiece $cnt]
   $uloader set upiece $upiece

}



set pst(prevfree) 0
#the tracker: for rarest first download

proc tracker {} {
  global ns pst opt


  set ctime [$ns now]
  vputs " in tracker (free $pst(free) prevfree $pst(prevfree))"

  set uplist [reshuffle $pst(dlist)] 
  foreach uploader $uplist {
	set ualloc [$uploader set ualloc]
	if { $ualloc >= $opt(U) } { continue }	
	set ptype [$uploader set ptype]
	if { $ptype == "SD" } { continue }	
	set maxd [$uploader set downloaded]

	# set dnlist [progsort $pst(dlist)]
	set dnlist [reshuffle $pst(dlist)]
	# set dnlist [reqsort $pst(dlist)]
	foreach downloader $dnlist {
		set dalloc [$downloader set dalloc]
		if {$dalloc >= $opt(D) } { continue}
		set ptype [$downloader set ptype]
		if { $ptype == "SD" || $ptype == "SS" || $ptype == "S"} {continue}
		set progress [$downloader set progress]
		if {$progress == $opt(M) } { continue}
		if {$progress > $opt(M) } { puts "high progress"; continue}
		
		# if { $progress >= $maxd } {continue}
		if { $ualloc >= $opt(U) } {continue}

		# select the target piece here
		set updpieces [$uploader set dpiece]
		set dndpieces [$downloader set dpiece]

		set tpiece -1
		set pieces [sortindex $opt(piececount)]
		foreach pindex $pieces {
			set us [lindex $updpieces $pindex]
			set ds [lindex $dndpieces $pindex]
			
			if { $us == "c" && $ds == "0" } {
				set tpiece $pindex
				break
			}
		}
		if { $tpiece < 1 } { continue }


		incr progress 1
		$downloader set progress $progress
		
		incr ualloc 
		$uploader set ualloc $ualloc

		incr dalloc
		$downloader set dalloc $dalloc

		
		set dndpieces [lreplace $dndpieces $tpiece $tpiece i]
		$downloader set dpiece $dndpieces
		
		
		set ctime [expr $ctime + 0.001]
		$ns at $ctime "startdownload $downloader $uploader $tpiece"
		vputs "  scheduling [$downloader set peerno] [$uploader set peerno] $tpiece"


		incr pst(demand) -1
		incr pst(free) -1

		if { $pst(free) == 0 } {
			set pst(prevfree) 0
			return 0
		}

	}

  }


   if { $pst(prevfree) == $pst(free) } { 
	vputs "calling validate"
	validatetracker
	return 0 
   }
   set pst(prevfree) $pst(free)

   if { $pst(free) > 0 && $pst(demand)>0 } {
	$ns at [expr $ctime + 0.050] "tracker"
   }	

}


proc validatetracker {} {
   global ns opt pst

   set ctime [$ns now]
   vputs " in validate tracker "

   lappend faults "0 0 0"
   foreach uploader $pst(dlist) {
	set uno [$uploader set peerno]
	set ualloc [$uploader set ualloc]
	if { $ualloc >= $opt(U) } {continue}
	set downloaded [$uploader set downloaded]
	if { $downloaded < 1 } { continue}
	set ptype [$uploader set ptype]	
	if { $ptype == "S" || $ptype == "SS" || $ptype == "SD" } {continue}

	

	foreach downloader $pst(dlist) {
		set dno [$downloader set peerno]
		if { $dno == $uno } { continue }
		set dalloc [$downloader set dalloc]
		if {$dalloc >= $opt(D) } {continue}
		set progress [$downloader set progress]
		if {$progress >= $opt(M) } {continue}
		set downloaded [$downloader set downloaded]
		if {$downloaded >= $opt(M) } {continue}
		
		set updpiece [$uploader set dpiece]		
		set dndpiece [$downloader set dpiece]

		for { set i 1 } { $i <= $opt(M) } { incr i} {
			set us [lindex $updpiece $i]
			set ds [lindex $dndpiece $i]
			if { $us == "c" && $ds == "0"} {
				lappend faults "$dno $uno $i"

			puts "$ctime validating :  tracker fails with piece $i"
			puts " validating U $uno / ualloc $ualloc: $updpiece"
			puts " validating D $dno / dalloc $dalloc: $dndpiece"
			puts ""
#			exit 36367875
			}
		}		
	}	
   } 
   vputs "$ctime from validate tracker "
   needputs "faults: [llength $faults] : $faults"
#   $ns at [expr 9.0 + $ctime ] "validatetracker"
}
#$ns at 9.0 validatetracker



set pst(sdcnt) 0
proc genseed {sdcnt} {
  global ns pst opt


  incr pst(sdcnt)
  set peerd [new PDATA $pst(sdcnt)]
  lappend pst(dlist) $peerd
  puts "[$ns now] creating seed $pst(sdcnt)"

 
  $peerd set ptype "SS"
  $peerd set downloaded $opt(M) 
  $peerd set dalloc $opt(D)
  $peerd set progress $opt(M)
  incr pst(free) $opt(U) 
  
  for { set i 0 } { $i <= $opt(M) } { incr i } {
	lappend dpiece c
  }
  $peerd set dpiece $dpiece	  


  set tcnt [expr $sdcnt - 1]
  if { $tcnt > 0 } {
     genseed $sdcnt
  }
}
$ns at 0.001 "genseed $opt(STSDCNT)"



#artificial: generates artificial seeds
proc artgenseed {sdcnt} {
  global ns pst opt


  incr pst(sdcnt) 1
  set peerd [new PDATA $pst(sdcnt)]
  lappend pst(dlist) $peerd
  puts "[$ns now] creating artificial seed $pst(sdcnt)"

 
  $peerd set ptype "S"
  $peerd set downloaded $opt(M) 
  $peerd set dalloc $opt(D)
  incr pst(free) $opt(U) 
  $peerd set progress $opt(M)

  for { set i 0 } { $i <= $opt(M) } { incr i } {
	lappend dpiece c
  }
  $peerd set dpiece $dpiece	  



  $ns at $opt(artdltime) "seeddepart $peerd"

  set tcnt [expr $sdcnt - 1]
  if { $tcnt > 0 } {
     artgenseed $tcnt
  }
}
if { $opt(artseedcnt) > 0  } {
   $ns at 0.002 "artgenseed $opt(artseedcnt)"
}



proc initpeerdata {peerno} {
  global ns pst opt

  puts "peer $peerno arrived"

  set peerd [new PDATA $peerno]
  lappend pst(dlist) $peerd
  
  set ctime [$ns now]
  $peerd set reqtime $ctime
  $peerd set ptype "D"        ;# for downloader
  $peerd set arrivaltime [$ns now]
   
  incr pst(dlcnt) 1

  incr pst(demand) $opt(D)
  if { $pst(free) > 0 && $pst(demand) > 0} {
       theputs "tracker called by initpeer"	
#       tracker      
  } 

}


proc peerarrival {} {
    global ns pst opt

    set curtime [$ns now]
    incr pst(peerno) 1
    initpeerdata $pst(peerno) 

    set rcomp [expr rand() * $opt(INTVAL) ]
    set rcomp [expr $rcomp * 0.25]
    set rcomp [expr $rcomp + $opt(INTVAL) ]
#    set rcomp $opt(INTVAL)
    $ns at [expr $curtime + $rcomp] "peerarrival"
}
set pst(peerno) [expr $opt(STSDCNT) + $pst(peerno) ]
set pst(peerno) [expr $pst(peerno) + $opt(artseedcnt)]
$ns at 0.003 peerarrival



set tfile [open "output.txt" w]  
proc trace {} {
   global ns opt pst tfile


   set ctime [$ns now]

   set dloader [lindex $pst(dlist) 1]
   set downloaded [$dloader set downloaded]

   set tsystime [clock seconds]
   set systime [clock format $tsystime -format %H:%M:%S ]
   set sysdate [clock format $tsystime -format %d:%m:%y ]

   puts ""
   set toutput [format "%0.2f  $pst(dlcnt) $pst(sdcnt)  M $opt(M) D $opt(D) U $opt(U) INT $opt(INTVAL) SDRES $opt(SDRES) DLTIME $opt(DLTIME) free $pst(free) demand $pst(demand) activeu $pst(activeu) time $systime $sysdate" $ctime ]
   puts $tfile $toutput
   puts $toutput


   # puts "Piece count :"
   # puts $opt(piececount)


    # matrix : upload/download matrix
    set itime [expr int($ctime)]
    set itime [expr $itime % 10]

set itime 3.0
if { $itime == 0 } {   
   puts "Piece uploading"
   foreach peer $pst(dlist) {
      puts "[$peer set peerno] : [$peer set ptype] / dd [$peer set downloaded] / activeu[$peer set activeu] / actived [$peer set actived]/[$peer set progress]"
      puts "[$peer set peerno]/ualloc [$peer set ualloc]/[$peer set ptype]/dd [$peer set downloaded] [$peer set upiece]"
   }     
    puts "Piece downloading"
   foreach peer $pst(dlist) {
      puts "[$peer set peerno][$peer set ptype]/dd[$peer set downloaded]:[$peer set dpiece]"
   }     

}

   set ctime [expr $ctime + $opt(TRINT)]
   $ns at $ctime trace
#   puts ""  
}
$ns at 1.0 trace



proc finish {} {
   global ns opt pst tfile
   flush $tfile
   close $tfile
   exit 0
}
$ns at 5000.0 finish


# tracks the number of pieces downloaded
for {set i 0} { $i <= $opt(M)} {incr i} { lappend opt(piececount) 0 }


$ns at 0.050 "tracker"

$ns run

